
Script.Load("lua/Utility.lua")
Script.Load("lua/Alien.lua")
Script.Load("lua/Mixins/BaseMoveMixin.lua")
Script.Load("lua/Mixins/GroundMoveMixin.lua")
Script.Load("lua/Mixins/CrouchMoveMixin.lua")
Script.Load("lua/Mixins/JumpMoveMixin.lua")
Script.Load("lua/CelerityMixin.lua")
Script.Load("lua/Mixins/CameraHolderMixin.lua")
--Script.Load("lua/WallMovementMixin.lua")
Script.Load("lua/DissolveMixin.lua")
--Script.Load("lua/BabblerClingMixin.lua")
Script.Load("lua/TunnelUserMixin.lua")
--Script.Load("lua/RailgunTargetMixin.lua")

Script.Load("lua/Prowler/VolleyRappel.lua")
Script.Load("lua/Prowler/AcidSpray.lua")
Script.Load("lua/Prowler/ReadyRoomRappel.lua")
--Script.Load("lua/Prowler/XenocideHowl.lua")
Script.Load("lua/Weapons/PredictedProjectile.lua")
Script.Load("lua/IdleMixin.lua")
--Script.Load("lua/SkulkVariantMixin.lua")

class 'Prowler' (Alien)

Prowler.kMapName = "prowler"

Prowler.kMaxSpeed = 6.6 -- skulk is 7.25, but 8.25 in Nin's Balance Mod
Prowler.kWalkBackwardSpeedScalar = 1.0
Prowler.kHorizontalJumpForce = 2.0 --7.2
Prowler.kVerticalExtraJumpForce = 3.0
Prowler.kSneakSpeedModifier = 0.7
Prowler.kMaxVerticalSpeed = 14 --8.5
Prowler.kRappelVerticalForce = 10
Prowler.kSilentSneakSpeed = 6.0
Prowler.kAirAddAcceleration = 1.0
Prowler.GlideAirSpeed = 8.5
Prowler.MaxAirSpeed = 11.5

Prowler.kHealth = kProwlerHealth
Prowler.kArmor  = kProwlerArmor

local is = 0
local kProwlerScale = 1.2
local kModelYScale = 1.3
local kProwlerVertAdjust = 0.1 * kProwlerScale
local kProwlerForwardAdjust = -0.15
local kProwlerAttackVertAdjust = 0.25
local kMass = 40
local kRappelDuration = 2.99  --not used

Prowler.kModelName = PrecacheAsset("models/prowler.model")
Prowler.kAnimationGraph = PrecacheAsset("models/prowler.animation_graph")
local kViewModelName = PrecacheAsset("models/prowler_view.model")
local kSkulkAnimationGraph = PrecacheAsset("models/prowler.animation_graph")
local kStepSound = PrecacheAsset("sound/NS2.fev/alien/skulk/footstep_right")

Prowler.kXExtents = .45 -- * kProwlerScale
Prowler.kYExtents = .45 -- * kProwlerScale
Prowler.kZExtents = .45 -- * kProwlerScale
Prowler.kViewOffsetHeight = .4 * kProwlerScale

Prowler.kHowlCooldown = 15 -- seconds
Prowler.kHowlEnergyCost = 45

Prowler.kGlideEnergyCost = 15 -- not used
Prowler.kHoverEnergyCost = 15

Prowler.kGravity = -21.5
Prowler.kGlideGravity = -13.5
Prowler.kRappelGravity = -4.3

if Server then
    Script.Load("lua/Prowler/Prowler_Server.lua", true)
elseif Client then
    Script.Load("lua/Prowler/Prowler_Client.lua", true)
end

local networkVars =
{
    --timeOfLastHowl = "private compensated time",
    
    gliding = "compensated boolean",
    rappelling = "compensated boolean",
    timeRappelStart = "private compensated time",
}

AddMixinNetworkVars(BaseMoveMixin, networkVars)
AddMixinNetworkVars(GroundMoveMixin, networkVars)
AddMixinNetworkVars(JumpMoveMixin, networkVars)
AddMixinNetworkVars(CrouchMoveMixin, networkVars)
AddMixinNetworkVars(CelerityMixin, networkVars)
AddMixinNetworkVars(CameraHolderMixin, networkVars)
--AddMixinNetworkVars( WallMovementMixin, networkVars)
AddMixinNetworkVars(DissolveMixin, networkVars)
--AddMixinNetworkVars(BabblerClingMixin, networkVars)
AddMixinNetworkVars(TunnelUserMixin, networkVars)
AddMixinNetworkVars(IdleMixin, networkVars)


function Prowler:OnCreate()
    

    InitMixin(self, BaseMoveMixin, { kGravity = Player.kGravity })
    InitMixin(self, GroundMoveMixin)
    InitMixin(self, JumpMoveMixin)
    InitMixin(self, CrouchMoveMixin)
    InitMixin(self, CelerityMixin)
    InitMixin(self, CameraHolderMixin, { kFov = kProwlerFov })
    --InitMixin(self, WallMovementMixin)

    Alien.OnCreate(self)

    InitMixin(self, DissolveMixin)
    --InitMixin(self, BabblerClingMixin)
    InitMixin(self, TunnelUserMixin)
    InitMixin(self, PredictedProjectileShooterMixin)
    
    if Client then
        --InitMixin(self, RailgunTargetMixin)
        self.timeDashChanged = 0
    end
    
    --self.timeOfLastHowl = 0
    --self.variant = kDefaultSkulkVariant
    
    self.runDist = 1
    self.gliding = false
    self.rappelling = false
    self.timeRappelStart = 0
    
end

function Prowler:GetPlayInstantRagdoll()
    return true
end

function Prowler:GetMapBlipType()
    return kMinimapBlipType.Skulk
end

function Prowler:OnInitialized()

    Alien.OnInitialized(self)
    
    
    self:SetModel(Prowler.kModelName, Prowler.kAnimationGraph)
    
    if Client then
    
        self.currentCameraRoll = 0
        self.goalCameraRoll = 0
        
        self:AddHelpWidget("GUIEvolveHelp", 2)
        self:AddHelpWidget("GUIMapHelp", 1)
        self:AddHelpWidget("GUITunnelEntranceHelp", 1)
        self:AddHelpWidget("GUIProwlerRappelHelp", 2)
        
    end

    InitMixin(self, IdleMixin)
    
end

function Prowler:GetViewModelName()
    return kViewModelName
end

-- these two are for third person mod
function Prowler:GetThirdPersonOffset()
  local z = -1.8 - self:GetVelocityLength() / self:GetMaxSpeed(true) * 0.4
  return Vector(0, 0.6, z) 
end

function Prowler:GetHeartOffset()
    return Vector(0, 0.6, 0)
end

function Prowler:GetFirstPersonFov()
  return kProwlerFov
end

function Prowler:GetStepLength()
    return 0.5
end

function Prowler:GetAirControl()
    return self.rappelling and 5 or 20 -- skulk is 27
end

function Prowler:GetAirAcceleration()
    return self.rappelling and 6.6 or 16 -- skulk is 9
end

function Prowler:GetCollisionSlowdownFraction()
    return 0.15
end
function Prowler:GetGroundTransistionTime()
    return 0.15
end

function Prowler:GetAngleSmoothRate()
    return 6
end

function Prowler:GetCollisionSlowdownFraction()
    return 0.1
end

function Prowler:GetRollSmoothRate()
    return 4
end

function Prowler:GetPitchSmoothRate()
    return 4
end

function Prowler:GetIsSmallTarget()
    return true
end

function Prowler:GetGroundFriction()
    return 11
end

function Prowler:GetAirFriction()
    return 0.06 -- - (GetHasCelerityUpgrade(self) and GetSpurLevel(self:GetTeamNumber()) or 0) * 0.006
end

local kProwlerEngageOffset = Vector(0, 0.5, 0)
function Prowler:GetEngagementPointOverride()
    return self:GetOrigin() + kProwlerEngageOffset
end

-- we, uh, don't have variants 
function Prowler:SetVariant()
end
function Prowler:GetVariant()
end
function Prowler:GetCrouchShrinkAmount()
    return 0
end
function Prowler:GetExtentsCrouchShrinkAmount()
    return 0
end
function Prowler:GetCrouchSpeedScalar()
    return 0
end

-- Tilt the camera based on the wall the Prowler is attached to.
function Prowler:PlayerCameraCoordsAdjustment(cameraCoords)

    local viewModelTiltAngles = Angles()
    viewModelTiltAngles:BuildFromCoords(Alien.PlayerCameraCoordsAdjustment(self, cameraCoords))

    if self.currentCameraRoll then
        viewModelTiltAngles.roll = viewModelTiltAngles.roll + self.currentCameraRoll
    end

    local viewModelTiltCoords = viewModelTiltAngles:GetCoords()
    viewModelTiltCoords.origin = cameraCoords.origin

    return viewModelTiltCoords

end

function Prowler:GetMaxSpeed(possible)

    if possible then
        return Prowler.GlideAirSpeed
    end
    
    local maxspeed = self:GetIsOnGround() and Prowler.kMaxSpeed or Prowler.GlideAirSpeed
    
    if self.movementModiferState then
        maxspeed = maxspeed * Prowler.kSneakSpeedModifier
    end
    
    return maxspeed
    
end

function Prowler:GetMaxBackwardSpeedScalar()
    return Prowler.kWalkBackwardSpeedScalar
end

function Prowler:OnProcessMove(input)

    Alien.OnProcessMove(self, input)
    
    if self:GetPlayFootsteps() and Client then
        is = is + 1
        if is > 1 then
            Shared.PlaySound(self, kStepSound, 1)
            self.playingDashSound = true
            is = 0
        end
    end

end

function Prowler:GetMaxViewOffsetHeight()
    return Prowler.kViewOffsetHeight
end

function Prowler:ModifyJump(input, velocity, jumpVelocity)
    -- add extra jump force when aiming upward
                                                          
    local wishDir

    if input.move:GetLength() > 0 then
        local notRecentlyLanded = self:GetTimeGroundTouched() + self:GetGroundTransistionTime() < Shared.GetTime()
        wishDir = self:GetViewCoords():TransformVector(input.move)
        local celerityBonus = GetHasCelerityUpgrade(self) and self:GetSpurLevel() * 0.4 or 0
        local verticalForce = wishDir.y * Prowler.kVerticalExtraJumpForce + celerityBonus
        wishDir.y = 0
        wishDir:Normalize()
        local jumpAddForce = notRecentlyLanded and (Prowler.kHorizontalJumpForce + celerityBonus) or 0.4
        jumpVelocity.y = jumpVelocity.y + verticalForce
        jumpVelocity:Add(wishDir * jumpAddForce)

    end                                       
    
end

function Prowler:OnJump( modifiedVelocity )

    local material = self:GetMaterialBelowPlayer()    
        
    self:TriggerEffects("jump", {surface = material})
    
end

function Prowler:GetIsRappelling( )
    return self.rappelling
end

function Prowler:ModifyCelerityBonus(celerityBonus)
    
    if self.movementModiferState then
        celerityBonus = celerityBonus * Prowler.kSneakSpeedModifier
    end
    
    return celerityBonus
    
end

function Prowler:ModifyGravityForce(gravityTable)
    if self:GetIsOnGround() then
        gravityTable.gravity = 0
    else
        --gravityTable.gravity = self:GetCrouching() and Prowler.kGravity or Prowler.kGlideGravity
        gravityTable.gravity = self.rappelling and Prowler.kRappelGravity or self.gliding and Prowler.kGlideGravity or Prowler.kGravity
    end
end

--local oldOnAdjustModelCoords = Prowler.OnAdjustModelCoords
function Prowler:OnAdjustModelCoords(modelCoords)
    --modelCoords = oldOnAdjustModelCoords(self, modelCoords)
    modelCoords.xAxis = modelCoords.xAxis * kProwlerScale
    modelCoords.yAxis = modelCoords.yAxis * kModelYScale
    modelCoords.zAxis = modelCoords.zAxis * kProwlerScale
    --modelCoords.origin.y = modelCoords.origin.y + kProwlerVertAdjust
    --[[if self.primaryAttacking then
        modelCoords.origin.y = modelCoords.origin.y + kProwlerAttackVertAdjust
    end--]]
    return modelCoords
end

function Prowler:GetDesiredAngles(deltaTime)

    local desiredAngles = Angles()
    if self.onGround then
        desiredAngles.pitch = 0
    else
        desiredAngles.pitch = self:GetIsJumping() and -0.3 or 0
    end
    desiredAngles.roll = self.viewRoll
    desiredAngles.yaw = self.viewYaw
    
    return desiredAngles

end

function Prowler:OnWorldCollision(normal)

    PROFILE("Prowler:OnWorldCollision")
    
    --self.rappelling = self.rappelling and normal.y < 0.7 --normal.y < 0.7
    
end

function Prowler:PreUpdateMove(input, runningPrediction)

    PROFILE("Prowler:PreUpdateMove")

    local notRecentlyJumped = Shared.GetTime() - self.timeOfLastJump > 0.09
    local glideDesired = bit.band(input.commands, Move.Jump) ~= 0 and notRecentlyJumped and not self:GetIsOnGround()
    if self.gliding ~= glideDesired then
        self.gliding = glideDesired
    end
    
    --[[if self.timeRappelStart + kRappelDuration < Shared.GetTime() then
        self.rappelling = false
    end--]]
end

--[[function Prowler:ModifyVelocity(input, velocity, deltaTime)

    if not self:GetIsOnGround() then --self:GetCrouching() and 
        -- slow fall
        local verticalForce = self:GetCrouching() and 0 or math.max(-velocity.y, 0)
        local celerityLevel = GetHasCelerityUpgrade(self) and self:GetSpurLevel() or 0
        local maxYSpeed = Prowler.kMaxVerticalSpeed + celerityLevel * 1.5
        -- glide in the desired direction while holding jump
        local energyCost = Prowler.kGlideEnergyCost
        if bit.band(input.commands, Move.Jump) ~= 0 then
            
            --if input.move:GetLength() > 0 then
                --verticalForce = verticalForce + 30 * Clamp(self:GetViewAngles():GetCoords().zAxis.y * 2, -1, 1) - velocity.y * 3
                velocity.y = math.min(velocity.y + verticalForce * deltaTime, maxYSpeed)
                
                local wishDir = self:GetViewCoords():TransformVector(input.move)
                wishDir.y = 0
                wishDir:Normalize()  -- hmmm
                
                local maxSpeed = Prowler.MaxAirSpeed + celerityLevel * 1.5
                if velocity:GetLengthXZ() < maxSpeed then
                    velocity:Add(wishDir * (Prowler.kAirAddAcceleration + celerityLevel)* deltaTime)
                end

                self:DeductAbilityEnergy(energyCost * deltaTime)
                
            --else
                --energyCost = Prowler.kHoverEnergyCost
            --    local speedScalar = (velocity:GetLength() < 1) and 0 or (1 - 3 * deltaTime)
            --    velocity.x = speedScalar * velocity.x
            --    velocity.y = (velocity.y > 0 and velocity.y or (speedScalar * velocity.y)) --+ 6 * deltaTime
            --    velocity.z = speedScalar * velocity.z
            end
            
            
            
    --elseif self.gliding and (velocity.y < 0) then
    --        verticalForce = -velocity.y
    --        velocity.y = math.min(velocity.y + verticalForce * deltaTime, maxYSpeed
    --    end
    end
    
end--]]

function Prowler:OnUpdateAnimationInput(modelMixin)

    PROFILE("Prowler:OnUpdateAnimationInput")
    
    Alien.OnUpdateAnimationInput(self, modelMixin)
end

function Prowler:GetPlayFootsteps()
    return self:GetVelocityLength() > Prowler.kSilentSneakSpeed and self:GetIsOnGround() and self:GetIsAlive() --and not self:GetCrouching()
end

function Prowler:GetMass()
    return kMass
end

function Prowler:GetBaseHealth()
    return Prowler.kHealth
end

function Prowler:GetBaseArmor()
    return Prowler.kArmor
end

function Prowler:GetHealthPerBioMass()
    return kProwlerHealthPerBioMass
end


function Prowler:SpawnCloud(techId)

    local position = self:GetOrigin() + self:GetViewCoords().zAxis * 5
    
    local mapName = LookupTechData(techId, kTechDataMapName)
    if mapName and Server then
    
        local cloudEntity = CreateEntity(mapName, position, self:GetTeamNumber())
        
    end

end

function Prowler:RappelMove()
    local velocity = self:GetVelocity()
    local forwardVec = self.rappelDirection
    local verticalVector = math.min(forwardVec.y + 0.2, 1)
    forwardVec.y = 0
    local newVelocity = velocity * 0.2 + forwardVec * 11.5 --10.3
    
    -- Add in vertical component.
    newVelocity.y = math.min(Prowler.kMaxVerticalSpeed, math.max(-Prowler.kMaxVerticalSpeed, velocity.y * 0.5) + Prowler.kRappelVerticalForce * verticalVector)
    
    self:SetVelocity(newVelocity)
    
    self.jumping = true
    self.rappelling = newVelocity.y >= 0 --true  -- rappel downward uses normal physics
    self:DisableGroundMove(0.2)
    self.timeRappelStart = Shared.GetTime()
    
    return false
end

function Prowler:OnRappel()

    self.rappelDirection = self:GetViewCoords().zAxis
    self:RappelMove()
    --[[if Server then
        self.rappelDirection = self:GetViewCoords().zAxis
        self:AddTimedCallback(self.RappelMove, 0.2)
    end--]]
    
end

function Prowler:OnHowl()

    --[[local shotWeb = false
    
    if GetHasTech(self, kTechId.ShiftHive, true) then
        self:SpawnCloud(kTechId.EnzymeCloud)
        shotWeb = true
    end
    if GetHasTech(self, kTechId.ShadeHive, true) then
        if Server then
            local newAlienExtents = LookupTechData(self:GetTechId(), kTechDataMaxExtents)
            local capsuleHeight, capsuleRadius = GetTraceCapsuleFromExtents(newAlienExtents) 
            
            local spawnPoint = GetRandomSpawnForCapsule(newAlienExtents.y, capsuleRadius, self:GetModelOrigin(), 0.5, 5)
            
            if spawnPoint then

                local hallucinatedPlayer = CreateEntity(Skulk.kMapName, spawnPoint, self:GetTeamNumber())
                
                hallucinatedPlayer:SetVariant(kSkulkVariant.normal)
                hallucinatedPlayer.isHallucination = true
                InitMixin(hallucinatedPlayer, PlayerHallucinationMixin)                
                InitMixin(hallucinatedPlayer, SoftTargetMixin)                
                InitMixin(hallucinatedPlayer, OrdersMixin, { kMoveOrderCompleteDistance = kPlayerMoveOrderCompleteDistance }) 

                hallucinatedPlayer:SetName(self:GetName())
                hallucinatedPlayer:SetHallucinatedClientIndex(self:GetClientIndex())
            
            end 
        end
        shotWeb = true
    end
    if GetHasTech(self, kTechId.CragHive, true) then
        self:SpawnCloud(kTechId.MucousMembrane)
        shotWeb = true
    end
    
    if shotWeb and Server then
        --self:TriggerEffects("drifter_shoot_enzyme", {effecthostcoords = self:GetCoords() } )
        self:TriggerEffects("drifter_shoot_enzyme", {effecthostcoords = Coords.GetLookIn(self:GetEyePos(), GetNormalizedVectorXZ(self:GetViewAngles():GetCoords().zAxis * 5 )) } )
    end--]]
    
end


if Server then
    function Prowler:OnKill(attacker, doer, point, direction)
    
        Alien.OnKill(self, attacker, doer, point, direction)
        --self:TriggerEffects("death", { classname = "Skulk", effecthostcoords = Coords.GetTranslation(self:GetOrigin()), doer = "Railgun"})
        
        local useModelName = self:GetModelName()
        local useGraphName = self:GetGraphName()
        
        local ragdoll = CreateEntity(Ragdoll.kMapName, self:GetOrigin())
        ragdoll:SetCoords(self:GetCoords())
        ragdoll:SetModel(useModelName, useGraphName)
        ragdoll:SetPhysicsType(PhysicsType.Dynamic)
        ragdoll:SetPhysicsGroup(PhysicsGroup.RagdollGroup)
        
        self:SetModel(nil)
    end
end

function Prowler:GetArmorFullyUpgradedAmount()
    return kProwlerArmorFullyUpgradedAmount
end

function Prowler:GetAnimateDeathCamera()
    return false
end

function Prowler:ModifyAttackSpeed(attackSpeedTable)

    local activeWeapon = self:GetActiveWeapon()
    if activeWeapon and activeWeapon:isa("AcidSpray") then
        attackSpeedTable.attackSpeed = attackSpeedTable.attackSpeed * 0.625
    end

end

Shared.LinkClassToMap("Prowler", Prowler.kMapName, networkVars, true)