
RappelMixin = CreateMixin( RappelMixin )
RappelMixin.type = "Rappel"

RappelMixin.overrideFunctions =
{
    "GetHasSecondary",
    "GetSecondaryEnergyCost",
--  "GetBarrelPoint",
    "PerformSecondaryAttack",
    "OnSecondaryAttack",
    "OnSecondaryAttackEnd"
}

kRappelRange = 20
local kRange = kRappelRange
local kRappelCooldown = 0.49
local kRappelDuration = 1.99
local kRappelProjectileSize = 0.3

RappelMixin.networkVars =
{
    rappelling                    = "private boolean",
    lastSecondaryAttackStartTime  = "private float",  -- todo: removed some networkvars if possible
    lastSecondaryAttackEndTime    = "private float",
}

function RappelMixin:__initmixin()
    
    PROFILE("RappelMixin:__initmixin")
    
    self.rappelling = false
    self.secondaryAttackingsecondaryAttacking = false
    self.lastSecondaryAttackStartTime = 0
    self.lastSecondaryAttackEndTime = 0

end

function RappelMixin:GetHasSecondary(player)
    return player.twoHives
end

function RappelMixin:GetSecondaryEnergyCost(player)
    return 20
end

function RappelMixin:EndRappel()
    local player = self:GetParent()
    if self.rappelling then
        self:OnSecondaryAttackEnd(self, player)
    end
    
    return false
end

function RappelMixin:EndRappelCallback()
    self:EndRappel()
end

function RappelMixin:PerformSecondaryAttack(player)
    
    local parent = self:GetParent()
    if parent and self:GetHasSecondary(player) and self.lastSecondaryAttackStartTime + kRappelCooldown < Shared.GetTime() then
        
        self.lastSecondaryAttackStartTime = Shared.GetTime()
        self.rappelling = true
        
        local viewAngles = player:GetViewAngles()
        local shootCoords = viewAngles:GetCoords()
        local filter = EntityFilterOneAndIsa(player, "Babbler")
        local startPoint = player:GetEyePos()
        local endPoint = startPoint + shootCoords.zAxis * kRange
        
        local extents = GetDirectedExtentsForDiameter(shootCoords.zAxis, kRappelProjectileSize)
        local trace = Shared.TraceBox(extents, startPoint, startPoint + shootCoords.zAxis * kRange, CollisionRep.Damage, PhysicsMask.Bullets, filter)
        
        if trace.fraction < 1 then
        
            local hitObject = trace.entity
            local direction = GetNormalizedVector(trace.endPoint - startPoint)
            local impactPoint = trace.endPoint - direction * kHitEffectOffset
            
            if hitObject and HasMixin(hitObject, "Team") and hitObject:GetTeamNumber() ~= self:GetTeamNumber() then
                self:DoDamage(20, hitObject, impactPoint, direction)
            end
            
            player:DeductAbilityEnergy(self:GetSecondaryEnergyCost())
            player:OnRappel()
            player:TriggerEffects("spikes_attack")
            return true
        end        
        
    end
    
    return false
    
end

function RappelMixin:OnSecondaryAttack(player)

    if player:GetEnergy() >= self:GetSecondaryEnergyCost() and not self.rappelling then
        self.secondaryAttacking = true
        self:PerformSecondaryAttack(player)
    end
    
    --[[if Server then
        self:AddTimedCallback(RappelMixin.EndRappel, kRappelDuration)
    end---]]
end

function RappelMixin:OnSecondaryAttackEnd(player)

    local now = Shared.GetTime()
    Ability.OnSecondaryAttackEnd(self, player)
    self.secondaryAttacking = false
    self.rappelling = false
    self.lastSecondaryAttackEndTime = now
    
end

function RappelMixin:OnHolster()
    
end

function RappelMixin:ProcessMoveOnWeapon(player, input)
    if not self.secondaryAttacking and player.rappelling and (player.timeRappelStart + 0.2 < Shared.GetTime()) then
        player.rappelling = false
    end
end

function RappelMixin:GetGhostModelName(ability)
    return Bomb.kModelName
end

if Client then


--[[    function RappelMixin:OnProcessIntermediate(input)

        local player = self:GetParent()
        local viewDirection = player:GetViewCoords().zAxis

        if player and self.rappelling then

            self.ghostCoords, self.placementValid = self:GetPositionForStructure(player:GetEyePos(), viewDirection, self:GetActiveStructure(), self.lastClickedPosition, self.lastClickedPositionNormal)

            if player:GetResources() < LookupTechData(self:GetActiveStructure():GetDropStructureId(), kTechDataCostKey) then
                self.placementValid = false
            end

        end

    end--]]
    
    function RappelMixin:GetGhostModelCoords()
    
        if self.showGhostModel then
            return self.ghostCoords
        end
        
    end
end