
Script.Load("lua/Weapons/Alien/Ability.lua")
Script.Load("lua/Prowler/RappelMixin.lua")


class 'VolleyRappel' (Ability)
VolleyRappel.kMapName = "volleyrappel"
VolleyRappel.kStartOffset = -1.25

local kAnimationGraph = PrecacheAsset("models/prowler_view.animation_graph")
local kVolleyRappelTracer = PrecacheAsset("cinematics/prowler/1p_tracer_residue.cinematic")
local kAttackDuration = Shared.GetAnimationLength("models/prowler_view.model", "bite_attack3") --* 0.746

-- higher numbers reduces the spread
local kSpreadDistance = 9.0
local kSpreadVertMult = 0.66
VolleyRappel.kSpreadVectors =
{
    GetNormalizedVector(Vector(0.0, 0.0, kSpreadDistance)),
    GetNormalizedVector(Vector(-0.05, 0.0, kSpreadDistance)),
    GetNormalizedVector(Vector( 0.05, 0.0, kSpreadDistance)),
    
    GetNormalizedVector(Vector(0.0, -0.3, kSpreadDistance)),
    GetNormalizedVector(Vector(0.0,  0.3, kSpreadDistance)),
    
    GetNormalizedVector(Vector(-0.7, -0.7, kSpreadDistance)),
    GetNormalizedVector(Vector(-0.7,  0.7, kSpreadDistance)),
    GetNormalizedVector(Vector( 0.7, -0.7, kSpreadDistance)),
    GetNormalizedVector(Vector( 0.7,  0.7, kSpreadDistance)),
    
    --GetNormalizedVector(Vector(-0.7, 0, kSpreadDistance)),
    --GetNormalizedVector(Vector(0.7, 0, kSpreadDistance)),
    --GetNormalizedVector(Vector(0, -0.7, kSpreadDistance)),
    --GetNormalizedVector(Vector(0, 0.7, kSpreadDistance)),
    
}

local networkVars =
{
}

AddMixinNetworkVars(RappelMixin, networkVars)

function VolleyRappel:OnCreate()

    Ability.OnCreate(self)
    InitMixin(self, RappelMixin)
    InitMixin(self, BulletsMixin)
    
    self.primaryAttacking = false

end
function VolleyRappel:GetAnimationGraphName()
    return kAnimationGraph
end
function VolleyRappel:GetVampiricLeechScalar()
    return kVolleyRappelVampirismScalar
end
function VolleyRappel:GetIsAffectedByFocus()
    return self.primaryAttacking
end
function VolleyRappel:GetAttackAnimationDuration()
    return kAttackDuration
end
function VolleyRappel:GetTracerEffectName()
    return kVolleyRappelTracer
end
function VolleyRappel:GetTracerResidueEffectName()
    return kVolleyRappelTracer
end
function VolleyRappel:GetBulletsPerShot()
    return 9
end

function VolleyRappel:GetEnergyCost(player)
    return kBiteEnergyCost
end
function VolleyRappel:GetHUDSlot()
    return 1
end
function VolleyRappel:GetRange()
    return 100
end
function VolleyRappel:GetBulletDamage()
    return kProwlerDamagePerPellet
end

function VolleyRappel:GetBarrelPoint()
    local player = self:GetParent()
    return player:GetEyePos() + Vector(0, -0.25, 0)
end

function VolleyRappel:GetDeathIconIndex()
    return kDeathMessageIcon.Spit
end

function VolleyRappel:GetDamageType()
    return kVolleyRappelDamageType
end

function VolleyRappel:OnPrimaryAttack(player)
    local hasEnergy = player:GetEnergy() >= self:GetEnergyCost()
    local cooledDown = (not self.nextAttackTime) or (Shared.GetTime() >= self.nextAttackTime)
    if hasEnergy and cooledDown then
        self.primaryAttacking = true
    else
        self.primaryAttacking = false
    end
    
end
function VolleyRappel:OnPrimaryAttackEnd()
    
    Ability.OnPrimaryAttackEnd(self)
    
    self.primaryAttacking = false
    
end
function VolleyRappel:OnUpdateAnimationInput(modelMixin)

    PROFILE("VolleyRappel:OnUpdateAnimationInput")
    
    local activityString = (self.primaryAttacking and "primary") or "none"
    modelMixin:SetAnimationInput("activity", activityString)
    
    local player = self:GetParent()
    if player then
        
        local viewmodel = player:GetViewModelEntity()
        if viewmodel  then
            viewmodel:SetIsVisible(true)
        end
    end
    
end

function VolleyRappel:OnTag(tagName)

    if tagName == "hit" then
        local player = self:GetParent()
        
        if player then
            local viewAngles = player:GetViewAngles()
            local roll = NetworkRandom() * math.pi * 2
            local rollAngles = Angles(0,0,roll):GetCoords()

            local shootCoords = viewAngles:GetCoords()
            shootCoords.yAxis = shootCoords.yAxis * kSpreadVertMult

            -- Filter ourself out of the trace so that we don't hit ourselves.
            local filter = EntityFilterTwo(player, self)
            local range = self:GetRange()
            
            local numberBullets = self:GetBulletsPerShot()
            local startPoint = player:GetEyePos()
            
            --self:TriggerEffects("shotgun_attack_sound")
            --self:TriggerEffects("shotgun_attack")
            
            for bullet = 1, math.min(numberBullets, #self.kSpreadVectors) do
            
                if not self.kSpreadVectors[bullet] then
                    break
                end
                
                local spreadVector = self.kSpreadVectors[bullet]
                spreadVector = rollAngles:TransformVector(spreadVector)
            
                local spreadDirection = shootCoords:TransformVector(spreadVector)

                local endPoint = startPoint + spreadDirection * range
                startPoint = player:GetEyePos() + shootCoords.xAxis * spreadVector.x * self.kStartOffset + shootCoords.yAxis * spreadVector.y * self.kStartOffset
                
                local targets, trace, hitPoints = GetBulletTargets(startPoint, endPoint, spreadDirection, 0.15, filter)
                
                local damage = 0

                HandleHitregAnalysis(player, startPoint, endPoint, trace)        
                    
                local direction = (trace.endPoint - startPoint):GetUnit()
                local hitOffset = direction * kHitEffectOffset
                local impactPoint = trace.endPoint - hitOffset
                local showTracer = true
                
                local numTargets = #targets
                
                if numTargets == 0 then
                    self:ApplyBulletGameplayEffects(player, nil, impactPoint, direction, 0, "rock", showTracer)
                end
                
                if Client and showTracer then
                    TriggerFirstPersonTracer(self, impactPoint)
                end
                
                for i = 1, numTargets do

                    local target = targets[i]
                    local hitPoint = hitPoints[i]

                    self:ApplyBulletGameplayEffects(player, target, hitPoint - hitOffset, direction, self:GetBulletDamage(), "rock", showTracer and i == numTargets)
                    
                    local client = Server and player:GetClient() or Client
                    if not Shared.GetIsRunningPrediction() and client and client.hitRegEnabled then
                        RegisterHitEvent(player, bullet, startPoint, trace, damage)
                    end
                
                end
                
            end
            self.shootingSpikes = true
            player:DeductAbilityEnergy(self:GetEnergyCost())
            if Server then
                self:TriggerEffects("drifter_parasite_hit")
            end
            
            self:DoAbilityFocusCooldown(player, kAttackDuration)
        end
    end

end

function VolleyRappel:GetSecondaryTechId()
    return kTechId.Rappel
end

Shared.LinkClassToMap("VolleyRappel", VolleyRappel.kMapName, networkVars)